package eu.mhsl.minecraft.pixelblocks.pixelblock;

import eu.mhsl.minecraft.pixelblocks.PixelBlocks;
import eu.mhsl.minecraft.pixelblocks.chunkGenerators.EmptyChunkGenerator;
import org.bukkit.*;
import org.bukkit.block.data.BlockData;
import org.bukkit.entity.*;
import org.bukkit.event.block.BlockBreakEvent;
import org.bukkit.event.block.BlockPlaceEvent;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.meta.ItemMeta;
import org.codehaus.plexus.util.FileUtils;

import java.io.File;
import java.io.IOException;
import java.util.*;

import static eu.mhsl.minecraft.pixelblocks.PixelBlocks.*;

public class PixelBlock {
    public static List<PixelBlock> placedPixelBlocks = new ArrayList<>();
    public static float hitboxOffset = 0.005F;
    public static int worldGrassBorderWidth = 5;
    public static int pixelsPerBlock = 16;
    public static boolean liveUpdate = true;

    public Location pixelBlockLocation;
    public ArrayList<Pixel> pixels = new ArrayList<>();
    public Interaction hitbox;
    public ItemDisplay barrier;

    public long lastEntryTime;
    public Location lastEntryLocation;
    public UUID ownerUID;
    public UUID uuid;

    public PixelBlock(Location originLocation, UUID ownerUID, UUID blockUUID) {
        this.uuid = blockUUID;

        this.pixelBlockLocation = originLocation.toBlockLocation();
        this.pixelBlockLocation.setYaw(0);
        this.pixelBlockLocation.setPitch(0);
        this.ownerUID = ownerUID;

        createPixelWorld();
    }

    public static PixelBlock getPixelBlockFromWorld(World world) {
        String blockUUID = Arrays.stream(world.getName().split(pathSeparator)).toList().getLast();
        List<PixelBlock> pixelBlocks = placedPixelBlocks.stream().filter(block -> block.uuid.toString().equals(blockUUID)).toList();

        if(!pixelBlocks.isEmpty()) {
            return pixelBlocks.getFirst();
        } else {
            return null;
        }
    }

    public static PixelBlock getPixelBlockFromLocation(Location location) {
        Location loc = location.clone().toBlockLocation();
        loc.setPitch(0);
        loc.setYaw(0);

        List<PixelBlock> pixelBlocks = placedPixelBlocks.stream()
            .filter(block -> block.pixelBlockLocation.equals(loc))
            .toList();

        if(pixelBlocks.isEmpty()) {
            return null;
        } else {
            return pixelBlocks.getFirst();
        }
    }

    public static String getWorldPathFromPixelblock(PixelBlock pixelBlock) {
        return plugin.getDataFolder().getPath() + pathSeparator + pixelBlock.uuid.toString();
    }


    void createPixelWorld() {
        File file = new File(getWorldPathFromPixelblock(this));
        final WorldCreator worldCreator = new WorldCreator(getWorldPathFromPixelblock(this));

        if(!file.exists() || !file.isDirectory()) {
            worldCreator.type(WorldType.FLAT);
            worldCreator.generator(new EmptyChunkGenerator());
            World newWorld = Bukkit.createWorld(worldCreator);

            assert newWorld != null;
            newWorld.setGameRule(GameRule.RANDOM_TICK_SPEED, 0);
            newWorld.setGameRule(GameRule.DO_FIRE_TICK, false);
            newWorld.setGameRule(GameRule.DO_MOB_SPAWNING, false);
            newWorld.setGameRule(GameRule.DO_WEATHER_CYCLE, false);
            newWorld.setGameRule(GameRule.DO_VINES_SPREAD, false);
            newWorld.setGameRule(GameRule.DO_DAYLIGHT_CYCLE, false);

            Location borderStartLocation = newWorld.getSpawnLocation().clone().subtract(1, 1, 1);
            Location grassStartLocation = borderStartLocation.clone().subtract(worldGrassBorderWidth, 0, worldGrassBorderWidth);

            Bukkit.getScheduler().runTask(plugin, () -> {
                for (int x = 0; x < (pixelsPerBlock+2) + 2 * worldGrassBorderWidth; x++) {
                    for (int z = 0; z < (pixelsPerBlock+2) + 2 * worldGrassBorderWidth; z++) {
                        grassStartLocation.clone().add(x, 0, z).getBlock().setType(Material.GRASS_BLOCK);
                    }
                }
                for (int x = 0; x < (pixelsPerBlock+2) + 2 * worldGrassBorderWidth; x++) {
                    for (int z = 0; z < (pixelsPerBlock+2) + 2 * worldGrassBorderWidth; z++) {
                        grassStartLocation.clone().add(x, -1, z).getBlock().setType(Material.DIRT);
                    }
                }
                for (int x = 0; x < (pixelsPerBlock+2); x++) {
                    for (int z = 0; z < (pixelsPerBlock+2); z++) {
                        Location currentLocation = borderStartLocation.clone().add(x, 0, z);

                        if (currentLocation.x() == borderStartLocation.x() || currentLocation.z() == borderStartLocation.z()) {
                            currentLocation.getBlock().setType(Material.RED_CONCRETE);
                        } else if (currentLocation.x() == borderStartLocation.x() + (pixelsPerBlock+1) || currentLocation.z() == borderStartLocation.z() + (pixelsPerBlock+1)) {
                            currentLocation.getBlock().setType(Material.RED_CONCRETE);
                        }
                    }
                }

                for (int x = (2*worldGrassBorderWidth+pixelsPerBlock-1)/2; x < (2*worldGrassBorderWidth+pixelsPerBlock-1)/2+4; x++) {
                    for (int y = 0; y < 5; y++) {
                        grassStartLocation.clone().add(x, 1+y, 0).getBlock().setType(Material.OBSIDIAN);
                    }
                }
                for (int x = (2*worldGrassBorderWidth+pixelsPerBlock-1)/2+1; x < (2*worldGrassBorderWidth+pixelsPerBlock-1)/2+3; x++) {
                    for (int y = 1; y < 4; y++) {
                        grassStartLocation.clone().add(x, 1+y, 0).getBlock().setType(Material.NETHER_PORTAL);
                    }
                }
            });
        } else {
            worldCreator.createWorld();
        }
    }

    public ItemStack getAsItem() {
        ItemStack itemStack = ItemStack.of(Material.GRAY_STAINED_GLASS);

        ItemMeta meta = itemStack.getItemMeta();
        meta.setDisplayName("Pixelblock von " + Objects.requireNonNull(Bukkit.getPlayer(this.ownerUID)).getName());
        meta.setEnchantmentGlintOverride(true);
        meta.setItemName(this.uuid.toString());
        itemStack.setItemMeta(meta);

        return itemStack;
    }

    public void handleInteraction(Player player) {
        if(!player.getUniqueId().equals(ownerUID)) {
            player.sendMessage("Dieser Pixelblock gehört nicht dir!");
            return;
        }

        this.lastEntryLocation = player.getLocation();
        this.lastEntryTime = System.currentTimeMillis();
        dataBase.savePixelBlock(this);

        player.teleport(getPlayerSpawnLocation(player));
    }

    public void handleAttack(Player player) {
//        if(!player.getUniqueId().equals(ownerUID)) {
//            player.sendMessage("Dieser Pixelblock gehört nicht dir!");
//            return;
//        }

        World world = Objects.requireNonNull(this.getPixelWorld());

        world.getPlayers().forEach(
                player1 -> {
                    player1.sendMessage("Der Pixelblock wurde abgebaut!");
                    player1.teleport(this.lastEntryLocation);
                }
        );

        world.getEntities().forEach(entity -> entity.teleport(this.lastEntryLocation));

        this.remove();
        this.pixelBlockLocation.getWorld().playSound(this.pixelBlockLocation, Sound.BLOCK_COPPER_BULB_BREAK, 1.0F, 30);
        this.pixelBlockLocation.getWorld().dropItem(this.pixelBlockLocation, this.getAsItem());
    }

    public void handleBlockBreak(BlockBreakEvent event) {
        Location blockLocation = event.getBlock().getLocation();
        Location spawnLocation = blockLocation.getWorld().getSpawnLocation();

        if(blockLocation.x() < spawnLocation.x() || blockLocation.z() < spawnLocation.z() || blockLocation.y() < spawnLocation.y()) {
            event.setCancelled(true);
            return;
        } else if(blockLocation.x() > spawnLocation.x()+(pixelsPerBlock-1) || blockLocation.z() > spawnLocation.z()+(pixelsPerBlock-1) || blockLocation.y() > spawnLocation.y()+(pixelsPerBlock-1)) {
            event.setCancelled(true);
            return;
        }

        Bukkit.getScheduler().runTask(plugin, () -> {
            if(liveUpdate) {
                Location relativeLocation = blockLocation.subtract(spawnLocation.getX(), spawnLocation.getY(), spawnLocation.getZ());
                relativeLocation.setWorld(this.pixelBlockLocation.getWorld());
                List<Pixel> pixels = this.pixels.stream().filter(pixel -> pixel.relativeLocation.equals(relativeLocation)).toList();
                if(!pixels.isEmpty()) {
                    Pixel pixel = pixels.getFirst();
                    pixel.remove();
                    this.pixels.remove(pixel);
                }
            }
        });
    }

    public void handleBlockPlace(BlockPlaceEvent event) {
        Location blockLocation = event.getBlock().getLocation();
        Location spawnLocation = blockLocation.getWorld().getSpawnLocation();

        if(blockLocation.x() < spawnLocation.x() || blockLocation.z() < spawnLocation.z() || blockLocation.y() < spawnLocation.y()) {
            event.setCancelled(true);
            return;
        } else if(blockLocation.x() > spawnLocation.x()+(pixelsPerBlock-1) || blockLocation.z() > spawnLocation.z()+(pixelsPerBlock-1) || blockLocation.y() > spawnLocation.y()+(pixelsPerBlock-1)) {
            event.setCancelled(true);
            return;
        }

        Bukkit.getScheduler().runTask(PixelBlocks.plugin, () -> {
            if(liveUpdate) {
                Location relativeLocation = blockLocation.subtract(spawnLocation.getX(), spawnLocation.getY(), spawnLocation.getZ());
                relativeLocation.setWorld(this.pixelBlockLocation.getWorld());

                Pixel newPixel = new Pixel(relativeLocation, event.getBlock().getBlockData(), ((double) 1 /pixelsPerBlock), 0);
                this.pixels.add(newPixel);
                newPixel.spawn(this.pixelBlockLocation);
            }
        });
    }

    public Location getPlayerSpawnLocation(Player player) {
        Location spawnLocation = getPixelWorld().getSpawnLocation().clone().add((double) pixelsPerBlock/2, 0, 1.5-worldGrassBorderWidth);
//        spawnLocation.setYaw(player.getLocation().getYaw());
//        spawnLocation.setPitch(player.getLocation().getPitch());
        return spawnLocation;
    }

    public Location getPortalLocation() {
        return getPixelWorld().getSpawnLocation().clone().add((double) pixelsPerBlock/2 -2, 0, -worldGrassBorderWidth-1);
    }

    public World getPixelWorld() {
        return Bukkit.getWorld(getWorldPathFromPixelblock(this));
    }

    public void spawnInteraction(boolean fullBlock) {
        if(fullBlock || pixels.isEmpty()) {
            hitbox = (Interaction) pixelBlockLocation.getWorld().spawnEntity(
                    pixelBlockLocation.clone().add(0.5, -hitboxOffset, 0.5),
                    EntityType.INTERACTION
            );
            hitbox.setInteractionHeight(1F + 2*hitboxOffset);
            hitbox.setInteractionWidth(1F + 2*hitboxOffset);
        } else {
            double startingX = this.pixels.stream()
                    .min(Comparator.comparing(pixel -> pixel.relativeLocation.getX()))
                    .orElseThrow().relativeLocation.getX();
            double startingY = this.pixels.stream()
                    .min(Comparator.comparing(pixel -> pixel.relativeLocation.getY()))
                    .orElseThrow().relativeLocation.getY();
            double startingZ = this.pixels.stream()
                    .min(Comparator.comparing(pixel -> pixel.relativeLocation.getZ()))
                    .orElseThrow().relativeLocation.getZ();

            double endingX = this.pixels.stream()
                    .max(Comparator.comparing(pixel -> pixel.relativeLocation.getX()))
                    .orElseThrow().relativeLocation.getX();
            double endingY = this.pixels.stream()
                    .max(Comparator.comparing(pixel -> pixel.relativeLocation.getY()))
                    .orElseThrow().relativeLocation.getY();
            double endingZ = this.pixels.stream()
                    .max(Comparator.comparing(pixel -> pixel.relativeLocation.getZ()))
                    .orElseThrow().relativeLocation.getZ();

            Location spawnLocation = pixelBlockLocation.clone().add(
                    ((startingX+endingX+1)/2)/pixelsPerBlock,
                    (startingY/pixelsPerBlock)-hitboxOffset,
                    ((startingZ+endingZ+1)/2)/pixelsPerBlock
                );

            float height = (float) (endingY-startingY+1)/pixelsPerBlock + 2*hitboxOffset;

            float width;
            if((endingX-startingX) > (endingZ-startingZ)) {
                width = (float) (endingX-startingX+1)/pixelsPerBlock + 2*hitboxOffset;
            } else {
                width = (float) (endingZ-startingZ+1)/pixelsPerBlock + 2*hitboxOffset;
            }

            if(spawnLocation.getX()+width/2 > this.pixelBlockLocation.getX()+1) {
                spawnLocation.subtract((spawnLocation.getX()+width/2)-(this.pixelBlockLocation.getX()+1), 0, 0);
            }
            if(spawnLocation.getX()-width/2 < this.pixelBlockLocation.getX()) {
                spawnLocation.add(this.pixelBlockLocation.getX()-(spawnLocation.getX()-width/2), 0, 0);
            }

            if(spawnLocation.getZ()+width/2 > this.pixelBlockLocation.getZ()+1) {
                spawnLocation.subtract(0, 0, (spawnLocation.getZ()+width/2)-(this.pixelBlockLocation.getZ()+1));
            }
            if(spawnLocation.getZ()-width/2 < this.pixelBlockLocation.getZ()) {
                spawnLocation.add(0, 0, this.pixelBlockLocation.getZ()-(spawnLocation.getZ()-width/2));
            }

            hitbox = (Interaction) pixelBlockLocation.getWorld().spawnEntity(
                spawnLocation,
                EntityType.INTERACTION
            );

            hitbox.setInteractionHeight(height);
            hitbox.setInteractionWidth(width);
        }
    }

    public void update() {
        Bukkit.getScheduler().runTask(plugin, () -> {
            this.clearEntities(true);

            for (int x = 0; x < pixelsPerBlock; x++) {
                for (int y = 0; y < pixelsPerBlock; y++) {
                    for (int z = 0; z < pixelsPerBlock; z++) {
                        Location relativeLocation = new Location(pixelBlockLocation.getWorld(), x, y, z);
                        Location blockLocation = Objects.requireNonNull(Bukkit
                                .getWorld(getWorldPathFromPixelblock(this)))
                                .getSpawnLocation()
                                .clone()
                                .add(relativeLocation.x(), relativeLocation.y(), relativeLocation.z());
                        BlockData block = blockLocation.getBlock().getBlockData();

                        if(block.getMaterial() != Material.AIR) {
                            Pixel newPixel = new Pixel(relativeLocation, block, ((double) 1 /pixelsPerBlock), 0);
                            pixels.add(newPixel);
                        }
                    }
                }
            }

            for(Pixel pixel : this.pixels) {
                pixel.spawn(this.pixelBlockLocation);
            }

            if(this.pixels.size() < 5) {
                Location relativeLocation = new Location(pixelBlockLocation.getWorld(), 0, 0, 0);
                BlockData block = Material.GRAY_STAINED_GLASS.createBlockData();
                Pixel newPixel = new Pixel(relativeLocation, block, 1, 0);
                pixels.add(newPixel);
                newPixel.spawn(this.pixelBlockLocation);

                Location itemDisplayLocation = this.pixelBlockLocation.clone().add(0.5, 0.5, 0.5);
                this.barrier = (ItemDisplay) this.pixelBlockLocation.getWorld().spawnEntity(
                        itemDisplayLocation,
                        EntityType.ITEM_DISPLAY
                );
                this.barrier.setItemStack(ItemStack.of(Material.BARRIER));

                spawnInteraction(true);
            } else {
                spawnInteraction(false);
            }
        });
    }

    public void clearEntities(boolean secure) {
        if(!this.pixelBlockLocation.getChunk().isEntitiesLoaded()) {
            this.pixelBlockLocation.getChunk().load(true);
            this.pixelBlockLocation.getChunk().getEntities();
        }

        if(!pixels.isEmpty()) {
            this.pixels.forEach(Pixel::remove);
            this.pixels.clear();
        }

        if(hitbox != null) {
            this.hitbox.remove();
            this.hitbox = null;
        }

        if(barrier != null) {
            this.barrier.remove();
            this.barrier = null;
        }

        if(secure) {
            List<Entity> entities = new ArrayList<>();
            entities.addAll(Bukkit.getWorlds().get(0).getEntities().stream()
                    .filter(entity -> entity.getType().equals(EntityType.BLOCK_DISPLAY) || entity.getType().equals(EntityType.INTERACTION) || entity.getType().equals(EntityType.ITEM_DISPLAY))
                    .toList());
            entities.addAll(Bukkit.getWorlds().get(1).getEntities().stream()
                    .filter(entity -> entity.getType().equals(EntityType.BLOCK_DISPLAY) || entity.getType().equals(EntityType.INTERACTION) || entity.getType().equals(EntityType.ITEM_DISPLAY))
                    .toList());
            entities.addAll(Bukkit.getWorlds().get(2).getEntities().stream()
                    .filter(entity -> entity.getType().equals(EntityType.BLOCK_DISPLAY) || entity.getType().equals(EntityType.INTERACTION) || entity.getType().equals(EntityType.ITEM_DISPLAY))
                    .toList());

            entities.stream().filter(entity -> entity.
                getLocation().
                clone().
                add(0, PixelBlock.hitboxOffset, 0).
                toBlockLocation().
                equals(this.pixelBlockLocation)).
                forEach(Entity::remove);
        }
    }

    public boolean place(Location placeLocation) {
        Location newLocation = placeLocation.toBlockLocation();
        newLocation.setPitch(0);
        newLocation.setYaw(0);

        if(PixelBlock.getPixelBlockFromLocation(newLocation) == null || PixelBlock.getPixelBlockFromLocation(newLocation) == this) {
            this.pixelBlockLocation = newLocation;
            update();
            dataBase.savePixelBlock(this);
            placedPixelBlocks.add(this);
            return true;
        }
        return false;
    }

    public void remove() {
        dataBase.removePixelBlock(this);

        clearEntities(true);
        placedPixelBlocks.remove(this);
    }

    public void delete() {
        this.remove();

        Bukkit.unloadWorld(getPixelWorld(), true);
        try {
            FileUtils.deleteDirectory(getWorldPathFromPixelblock(this));
        } catch (IOException e) {
            System.err.println("World could not be deleted: " + e.getMessage());
        }
    }
}
